/* 
 *  GGI/DM default mode setting
 *
 *  Copyright (c) 1999 Nicolai Haehnle - Locinga@pfaffenwinkel.de
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define DEBUG		1

#include <stdlib.h>
#include <string.h>

#include <ggi/internal/ggi-dl.h>
#include <ggi/internal/dm_int.h>
#include "default.h"


/*
 * Make the first drawmap in the flipping chain visible
 */
static int physically_set_visible(dmext *hook)
{
	ggidm_drawmap dm;
	
	GGIDPRINT("physically_set_visible called\n");
	
	dm = hook->fchain;
	
	GGIDPRINT("Setting origin of %p to (%i/%i)... ", dm->vis,
			dm->pos.x, dm->pos.y);
	
	if (ggiSetOrigin(dm->vis, dm->pos.x, dm->pos.y)) {
		GGIDPRINT("failed\n");
		return GGIDM_ERR_IO;
	}
	
	GGIDPRINT("OK\n");
	return 0;
}

/*
 * Cleans up previous mode (unacquire *all* drawmaps unless graphtype
 * stays the same)
 *	---
 * Create and insert visible drawmaps; establish flipping chain.
 * Acquire all previously created drawmaps.
 * Set the graphics mode physically
 *
 * WARNING:
 *  This destroys all visible drawmaps (unless the mode doesn't change,
 *  but don't rely on that).
 */

static int cleanup_mode(dmext *hook, int unacqall)
{
	dmvisual *dv;

	GGIDPRINT("cleanup_mode() called for %p\n", hook);

	/* Clear the old status */
	GGIDPRINT("Invalidating video memory drawmaps... ");
	dv = hook->dmvis;
	do {
		dmvisual *tv = dv;
		dv = dv->next;
		
		if ((tv->flags & GGIDM_VISIBLE) || unacqall) {
			while(tv->drawmaps) {
				if (__GGIDMacquire(hook, tv->drawmaps, 0)) {
					GGIDPRINT("Unacquire failed\n");
				}
			}
		}
		
	} while(dv != hook->dmvis);
	
	dv->flags &= ~GGIDM_ACQUIRED;
	dv->size.x = dv->size.y = 0;
	GGIDPRINT("OK\n");
	
	GGIDPRINT("Freeing freerects... ");
	while(dv->freerects) {
		dmrect *rc = dv->freerects;
		REMOVE_LIST(&dv->freerects, rc);
		free(rc);
	}
	GGIDPRINT("OK\n");
	
	GGIDPRINT("Freeing flipping chain... ");
	while(hook->fchain) {
		ggidm_drawmap dm;
		dm = hook->fchain;
		REMOVE_LIST(&hook->fchain, dm);
		free(dm);
	}
	GGIDPRINT("OK\n");
	
	return 0;
}

static int create_fchain(dmext *hook, dmvisual *dv)
{
	ggidm_drawmap dm;
	int i;

	GGIDPRINT("called create_fchain(%p, %p)\n", hook, dv);

	/* Create and insert visible drawmaps */
	for(i = 0; i < hook->frames; i++) {
		GGIDPRINT("Creating visible drawmap %i\n", i);
	
		dm = (ggidm_drawmap)malloc(sizeof(*dm));
		
		dm->hook = hook;
		dm->vis = hook->dmvis->vis;
		dm->parent = dv;
		
		dm->flags = GGIDM_VISIBLE | GGIDM_STORE_VIDEOMEMORY;
		dm->size.x = hook->visible.x;
		dm->size.y = hook->visible.y; 
		
		insert_drawmap(dv, dm);
		GGIDPRINT("placed at (%i/%i)\n", dm->pos.x, dm->pos.y);
		
		/* insert it into flipping chain */
		INSERT_LIST(&hook->fchain, dm);
	}
	
	GGIDPRINT("hook->fchain=%p (%i/%i), hook->fchain->next=%p (%i/%i)\n",
		hook->fchain, hook->fchain->pos.x, hook->fchain->pos.y,
		hook->fchain->next, hook->fchain->next->pos.x,
		hook->fchain->next->pos.y);
	
	GGIDPRINT("Done\n");
	
	return 0;
}

/*
 * Calls __GGIDMacquire for all drawmaps in list, with storage type store
 */
static int acquire_drawmaps(dmext *hook, int store)
{
	ggidm_drawmap dm;
	ggidm_drawmap *list = &hook->drawmaps;
	int error;

	GGIDPRINT("called acquire_drawmaps(%p, %i)\n", list, store);

 restart:
	dm = *list;
	if (dm) {
		GGIDPRINT("Beginning loop with %p\n", dm);
		do {
			GGIDPRINT("Checking %p\n", dm);
			if ((dm->flags & GGIDM_STORE_MASK) == store) {
				GGIDPRINT("Calling __GGIDMacquire(dm, 1)\n");
				error = __GGIDMacquire(hook, dm, 1);
				if (error) {
					GGIDPRINT("Failed\n");
					return error;
				}
				GGIDPRINT("Success\n");
				goto restart;
			}
			dm = dm->next;
		} while(dm != *list);
	}
	
	GGIDPRINT("Done\n");
	return 0;
}


int GGIDMsetmode(ggi_visual_t vis, int sizex, int sizey, int frames,
		 ggi_graphtype type)
{
	ggi_mode md;
	dmext *hook = LIBGGI_DMEXT(vis);
	dmvisual *dv;
	int error;

	GGIDPRINT("GGIDMsetmode: setting (%ix%ix%i) %i frames\n",
		sizex, sizey, GT_DEPTH(type), frames);

	/* some checks */
	if (hook->visible.x == sizex && hook->visible.y == sizey &&
	    hook->type == type && hook->frames == frames) {
	   	GGIDPRINT("Mode doesn't change -- doing nothing\n");
	   	return 0;
	}

	if (ggiCheckSimpleMode(vis, sizex, sizey, 1, type, &md)) {
		GGIDPRINT("checkmode failed\n");
		return GGIDM_ERR_INVALID;
	}
	
	dv = hook->dmvis;
	error = cleanup_mode(hook, hook->type != type);
	if (error)
		return error;
	
	/* Setups */
	hook->visible.x = sizex;
	hook->visible.y = sizey;
	hook->type = type;
	hook->frames = frames;
	
	error = create_fchain(hook, dv);
	if (error)
		return error;
	
	/* Acquire drawmaps */
	GGIDPRINT("Acquire previously allocated drawmaps\n");
	
	GGIDPRINT("vidmem...\n");
	error = acquire_drawmaps(hook, GGIDM_STORE_VIDEOMEMORY);
	if (error)
		return error;
	
 	GGIDPRINT("dontworry...\n");
	error = acquire_drawmaps(hook, GGIDM_STORE_DONTWORRY);
	if (error)
		return error;
	
	GGIDPRINT("and the rest...");
 	while(hook->drawmaps) {
 		error = __GGIDMacquire(hook, hook->drawmaps, 1);
 		if (error) {
 			GGIDPRINT("__GGIDMacquire() failed\n");
 			return error;
 		}
 	}
 	GGIDPRINT("OK\n");
 	
 	/* Physically set the mode */
 	GGIDPRINT("Set the mode physically\n");
 	dv = hook->dmvis;
 	do {
 		if (!(dv->flags & GGIDM_ACQUIRED))
	 		if (physically_set_mode(hook, dv))
	 			return GGIDM_ERR_IO;
 		dv = dv->next;
 	} while(dv != hook->dmvis);
	
	GGIDPRINT("Set visible visual origin\n");
	if (physically_set_visible(hook))
		return GGIDM_ERR_IO;
	
	return 0;
}


int GGIDMsetflags(ggi_visual_t vis, ggi_flags flags)
{
	dmext *hook = LIBGGI_DMEXT(vis);
	dmvisual *dv;
	
	GGIDPRINT("called GGIDMsetflags(%p, %08X)\n", vis, flags);
	
	if (flags == hook->flags) {
		GGIDPRINT("no change\n");
		return 0;
	}
	
	hook->flags = flags;
	dv = hook->dmvis;
	do {
		if (dv->flags & GGIDM_VISIBLE) {
			hook->old_setflags(vis, flags);
		} else {
			ggiSetFlags(dv->vis, flags);
		}
	
		dv = dv->next;
	} while(dv != hook->dmvis);
	
	return 0;
}

/*
 * NOTE: ggiFlushRegion doesn't make a lot of sense on the main visual, 
 *	 so we flush the complete visual here
 */
int GGIDMflush(ggi_visual_t vis, int x, int y, int w, int h, int tryflag)
{
	dmext *hook = LIBGGI_DMEXT(vis);
	dmvisual *dv;
	
	GGIDPRINT("called GGIDMflush(%p, %i)\n", vis);
	
	dv = hook->dmvis;
	do {
		if (dv->flags & GGIDM_VISIBLE)
			hook->old_flush(vis, 0, 0, LIBGGI_X(vis),
					LIBGGI_Y(vis), tryflag);
		else
			dv->vis->opdisplay->flush(dv->vis, 0, 0,
				LIBGGI_X(vis), LIBGGI_Y(vis), tryflag);
	
		dv = dv->next;
	} while(dv != hook->dmvis);
	
	return 0;
}

int GGIDMsetpalvec(ggi_visual_t vis, int start, int len, ggi_color *colormap)
{
	dmext *hook = LIBGGI_DMEXT(vis);
	dmvisual *dv;
    	ggi_color *newcm;
    	int newstart, newend, newlen;
	
	GGIDPRINT("called GGIDMsetpalvec(%p, %i, %i, %p)\n", 
		vis, start, len, colormap);
	
	if (start == -1)
		start = 0;
	
	newstart = hook->cm_start;
	if (newstart > start)
		newstart = start;
	newend = hook->cm_start + hook->cm_len;
	if (newend < start+len)
		newend = start + len;
	newlen = newend - newstart;
	
	newcm = hook->colormap;
	
	if (newstart != hook->cm_start  || newlen != hook->cm_len) {
	    	GGIDPRINT("Allocating new colormap\n");
	    	
	    	newcm = (ggi_color *)malloc(sizeof(*newcm)*newlen);
	    	memset(newcm, 0, sizeof(*newcm)*newlen);
	    	
	    	if (hook->colormap) {
		    	memcpy(&newcm[hook->cm_start - newstart], 
		    		hook->colormap,	sizeof(*newcm)*hook->cm_len);
		
			free(hook->colormap);
		}
		hook->colormap = newcm;
		hook->cm_start = newstart;
		hook->cm_len = newlen;
	}
	
	memcpy(&newcm[start - newstart], colormap, sizeof(*newcm)*len);
	
	dv = hook->dmvis;
	do {
		if (dv->flags & GGIDM_VISIBLE)
			hook->old_setpalvec(vis, newstart, newlen, 
				hook->colormap);
		else
			dv->vis->opcolor->setpalvec(dv->vis, newstart, 
				newlen, hook->colormap);
	
		dv = dv->next;
	} while(dv != hook->dmvis);
	
	return 0;
}

/*
 * Flips to the next drawmap in the flipping chain.
 * NOTE:
 *  Say you have a dmPrimary (pointing to the currently visible drawmap)
 *  and a dmBackbuffer (pointing to the backbuffer). After calling
 *  ggidmFlip() dmPrimary still points to the visible drawmap and
 *  dmBackbuffer to the backbuffer.
 */
int GGIDMflip(ggi_visual_t vis)
{
	ggi_coord save_pos;
	ggidm_drawmap dm;
	dmext *hook = LIBGGI_DMEXT(vis);
	
	GGIDPRINT("called ggidmFlip(%p)\n", vis);
	
	dm = hook->fchain;
	if (!dm) {
		GGIDPRINT("No mode set (fchain is empty)\n");
		return GGIDM_ERR_NOTACQUIRED;
	}
	memcpy(&save_pos, &dm->pos, sizeof(dm->pos));
	do {
		memcpy(&dm->pos, &dm->prev->pos, sizeof(dm->pos));
		dm = dm->prev;
	} while(dm != hook->fchain);
	
	memcpy(&hook->fchain->next->pos, &save_pos, sizeof(save_pos));
	
	GGIDPRINT("Physically set the new primary drawmap\n");
	return physically_set_visible(hook);
}


ggi_visual_t GGIDMgetvisual(ggidm_drawmap dm)
{
	ggi_visual_t subvis;

	if (!(dm->flags & GGIDM_ACQUIRED))
		return 0;
	dm->flags &= ~GGIDM_INVALID;

	subvis = ggiOpen("display-sub", dm->vis);
	ggiSetGraphMode(subvis, dm->pos.x, dm->pos.y, dm->size.x, dm->size.y,
		dm->hook->type);
	
	return subvis;
}
